@@ -14,6 +14,10 @@
 #include <stdlib.h>
 #include <string.h>
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #ifdef WEBP_HAVE_PNG
 #include <png.h>
 #endif
@@ -23,7 +27,10 @@
 #include <jpeglib.h>
 #endif
 
-#ifdef _WIN32
+#ifdef HAVE_WINCODEC_H
+#ifdef __MINGW32__
+#define INITGUID  // Without this GUIDs are declared extern and fail to link
+#endif
 #define CINTERFACE
 #define COBJMACROS
 #define _WIN32_IE 0x500  // Workaround bug in shlwapi.h when compiling C++
@@ -31,13 +38,26 @@
 #include <shlwapi.h>
 #include <windows.h>
 #include <wincodec.h>
+
+#ifndef GUID_WICPixelFormat24bppRGB
+// From Microsoft SDK 7.0a
+DEFINE_GUID(GUID_WICPixelFormat24bppRGB,
+    0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0d);
 #endif
+#ifndef GUID_WICPixelFormat32bppRGBA
+DEFINE_GUID(GUID_WICPixelFormat32bppRGBA,
+    0xf5c7ad2d, 0x6a8d, 0x43dd, 0xa7, 0xa8, 0xa2, 0x99, 0x35, 0x26, 0x1a, 0xe9);
+#endif
+#endif  /* HAVE_WINCODEC_H */
 
 
 #include "webp/encode.h"
 #include "stopwatch.h"
+#ifndef WEBP_DLL
+extern void* VP8GetCPUInfo;   // opaque forward declaration.
+#endif
 
-//-----------------------------------------------------------------------------
+//------------------------------------------------------------------------------
 
 static int verbose = 0;
 
@@ -68,7 +88,7 @@ static int ReadYUV(FILE* in_file, WebPPicture* const pic) {
   return ok;
 }
 
-#ifdef _WIN32
+#ifdef HAVE_WINCODEC_H
 
 #define IFS(fn)                \
   do {                         \
@@ -95,7 +115,7 @@ static HRESULT OpenInputStream(const char* filename, IStream** ppStream) {
 }
 
 static HRESULT ReadPictureWithWIC(const char* filename,
-                                  WebPPicture* const pic) {
+                                  WebPPicture* const pic, int keep_alpha) {
   HRESULT hr = S_OK;
   IWICBitmapFrameDecode* pFrame = NULL;
   IWICFormatConverter* pConverter = NULL;
@@ -105,6 +125,15 @@ static HRESULT ReadPictureWithWIC(const char* filename,
   UINT frameCount = 0;
   UINT width, height = 0;
   BYTE* rgb = NULL;
+  WICPixelFormatGUID srcPixelFormat = { 0 };
+  GUID srcContainerFormat = { 0 };
+  const GUID* alphaContainers[] = {
+    &GUID_ContainerFormatBmp,
+    &GUID_ContainerFormatPng,
+    &GUID_ContainerFormatTiff
+  };
+  int has_alpha = 0;
+  int i, stride;
 
   IFS(CoInitialize(NULL));
   IFS(CoCreateInstance(MAKE_REFGUID(CLSID_WICImagingFactory), NULL,
@@ -125,28 +154,53 @@ static HRESULT ReadPictureWithWIC(const char* filename,
     hr = E_FAIL;
   }
   IFS(IWICBitmapDecoder_GetFrame(pDecoder, 0, &pFrame));
+  IFS(IWICBitmapFrameDecode_GetPixelFormat(pFrame, &srcPixelFormat));
+  IFS(IWICBitmapDecoder_GetContainerFormat(pDecoder, &srcContainerFormat));
+
+  has_alpha = keep_alpha;
+  for (i = 0;
+       has_alpha && i < sizeof(alphaContainers)/sizeof(alphaContainers[0]);
+       ++i) {
+    if (IsEqualGUID(&srcContainerFormat, alphaContainers[i])) {
+      has_alpha =
+          IsEqualGUID(&srcPixelFormat, &GUID_WICPixelFormat32bppRGBA) ||
+          IsEqualGUID(&srcPixelFormat, &GUID_WICPixelFormat32bppBGRA);
+      break;
+    }
+  }
 
   // Prepare for pixel format conversion (if necessary).
   IFS(IWICImagingFactory_CreateFormatConverter(pFactory, &pConverter));
   IFS(IWICFormatConverter_Initialize(pConverter, (IWICBitmapSource*)pFrame,
-          MAKE_REFGUID(GUID_WICPixelFormat24bppRGB), WICBitmapDitherTypeNone,
+          has_alpha ? MAKE_REFGUID(GUID_WICPixelFormat32bppRGBA)
+                    : MAKE_REFGUID(GUID_WICPixelFormat24bppRGB),
+          WICBitmapDitherTypeNone,
           NULL, 0.0, WICBitmapPaletteTypeCustom));
 
   // Decode.
   IFS(IWICFormatConverter_GetSize(pConverter, &width, &height));
+  stride = (has_alpha ? 4 : 3) * width * sizeof(*rgb);
   if (SUCCEEDED(hr)) {
-    rgb = (BYTE*)malloc(3 * width * height);
+    rgb = (BYTE*)malloc(stride * height);
     if (rgb == NULL)
       hr = E_OUTOFMEMORY;
   }
-  IFS(IWICFormatConverter_CopyPixels(pConverter, NULL, 3 * width,
-          3 * width * height, rgb));
+  IFS(IWICFormatConverter_CopyPixels(pConverter, NULL, stride,
+          stride * height, rgb));
 
   // WebP conversion.
   if (SUCCEEDED(hr)) {
+    int ok;
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+    if (has_alpha) {
+      pic->colorspace |= WEBP_CSP_ALPHA_BIT;
+    }
+#endif
     pic->width = width;
     pic->height = height;
-    if (!WebPPictureImportRGB(pic, rgb, 3 * width))
+    ok = has_alpha ? WebPPictureImportRGBA(pic, rgb, stride)
+                   : WebPPictureImportRGB(pic, rgb, stride);
+    if (!ok)
       hr = E_FAIL;
   }
 
@@ -160,7 +214,8 @@ static HRESULT ReadPictureWithWIC(const char* filename,
   return hr;
 }
 
-static int ReadPicture(const char* const filename, WebPPicture* const pic) {
+static int ReadPicture(const char* const filename, WebPPicture* const pic,
+                       int keep_alpha) {
   int ok;
   if (pic->width != 0 && pic->height != 0) {
     // If image size is specified, infer it as YUV format.
@@ -173,7 +228,7 @@ static int ReadPicture(const char* const filename, WebPPicture* const pic) {
     fclose(in_file);
   } else {
     // If no size specified, try to decode it using WIC.
-    ok = SUCCEEDED(ReadPictureWithWIC(filename, pic));
+    ok = SUCCEEDED(ReadPictureWithWIC(filename, pic, keep_alpha));
   }
   if (!ok) {
     fprintf(stderr, "Error! Could not process file %s\n", filename);
@@ -181,7 +236,7 @@ static int ReadPicture(const char* const filename, WebPPicture* const pic) {
   return ok;
 }
 
-#else  // !_WIN32
+#else  // !HAVE_WINCODEC_H
 
 #ifdef WEBP_HAVE_JPEG
 struct my_error_mgr {
@@ -268,6 +323,8 @@ static int ReadJPEG(FILE* in_file, WebPPicture* const pic) {
 
 #else
 static int ReadJPEG(FILE* in_file, WebPPicture* const pic) {
+  (void)in_file;
+  (void)pic;
   printf("JPEG support not compiled. Please install the libjpeg development "
          "package before building.\n");
   return 0;
@@ -280,10 +337,11 @@ static void PNGAPI error_function(png_structp png, png_const_charp dummy) {
   longjmp(png_jmpbuf(png), 1);
 }
 
-static int ReadPNG(FILE* in_file, WebPPicture* const pic) {
+static int ReadPNG(FILE* in_file, WebPPicture* const pic, int keep_alpha) {
   png_structp png;
   png_infop info;
   int color_type, bit_depth, interlaced;
+  int has_alpha;
   int num_passes;
   int p;
   int ok = 0;
@@ -324,13 +382,24 @@ static int ReadPNG(FILE* in_file, WebPPicture* const pic) {
   }
   if (png_get_valid(png, info, PNG_INFO_tRNS)) {
     png_set_tRNS_to_alpha(png);
+    has_alpha = 1;
+  } else {
+    has_alpha = !!(color_type & PNG_COLOR_MASK_ALPHA);
   }
 
-  // TODO(skal): Strip Alpha for now (till Alpha is supported).
-  png_set_strip_alpha(png);
+  if (!keep_alpha) {
+    png_set_strip_alpha(png);
+    has_alpha = 0;
+  }
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+  if (has_alpha) {
+    pic->colorspace |= WEBP_CSP_ALPHA_BIT;
+  }
+#endif
+
   num_passes = png_set_interlace_handling(png);
   png_read_update_info(png, info);
-  stride = 3 * width * sizeof(*rgb);
+  stride = (has_alpha ? 4 : 3) * width * sizeof(*rgb);
   rgb = (uint8_t*)malloc(stride * height);
   if (rgb == NULL) goto Error;
   for (p = 0; p < num_passes; ++p) {
@@ -344,14 +413,18 @@ static int ReadPNG(FILE* in_file, WebPPicture* const pic) {
 
   pic->width = width;
   pic->height = height;
-  ok = WebPPictureImportRGB(pic, rgb, stride);
+  ok = has_alpha ? WebPPictureImportRGBA(pic, rgb, stride)
+                 : WebPPictureImportRGB(pic, rgb, stride);
   free(rgb);
 
  End:
   return ok;
 }
 #else
-static int ReadPNG(FILE* in_file, WebPPicture* const pic) {
+static int ReadPNG(FILE* in_file, WebPPicture* const pic, int keep_alpha) {
+  (void)in_file;
+  (void)pic;
+  (void)keep_alpha;
   printf("PNG support not compiled. Please install the libpng development "
          "package before building.\n");
   return 0;
@@ -383,7 +456,8 @@ static InputFileFormat GetImageType(FILE* in_file) {
   return format;
 }
 
-static int ReadPicture(const char* const filename, WebPPicture* const pic) {
+static int ReadPicture(const char* const filename, WebPPicture* const pic,
+                       int keep_alpha) {
   int ok = 0;
   FILE* in_file = fopen(filename, "rb");
   if (in_file == NULL) {
@@ -395,7 +469,7 @@ static int ReadPicture(const char* const filename, WebPPicture* const pic) {
     // If no size specified, try to decode it as PNG/JPEG (as appropriate).
     const InputFileFormat format = GetImageType(in_file);
     if (format == PNG) {
-      ok = ReadPNG(in_file, pic);
+      ok = ReadPNG(in_file, pic, keep_alpha);
     } else if (format == JPEG) {
       ok = ReadJPEG(in_file, pic);
     }
@@ -411,7 +485,7 @@ static int ReadPicture(const char* const filename, WebPPicture* const pic) {
   return ok;
 }
 
-#endif  // !_WIN32
+#endif  // !HAVE_WINCODEC_H
 
 static void AllocExtraInfo(WebPPicture* const pic) {
   const int mb_w = (pic->width + 15) / 16;
@@ -473,6 +547,14 @@ static void PrintExtraInfo(const WebPPicture* const pic, int short_output) {
               100.f * stats->header_bytes[0] / stats->coded_size,
               stats->header_bytes[1],
               100.f * stats->header_bytes[1] / stats->coded_size);
+      if (stats->alpha_data_size) {
+        fprintf(stderr, "             transparency:   %6d\n",
+                stats->alpha_data_size);
+      }
+      if (stats->layer_data_size) {
+        fprintf(stderr, "             enhancement:    %6d\n",
+                stats->layer_data_size);
+      }
       fprintf(stderr, " Residuals bytes  "
                       "|segment 1|segment 2|segment 3"
                       "|segment 4|  total\n");
@@ -519,7 +601,7 @@ static void PrintExtraInfo(const WebPPicture* const pic, int short_output) {
   }
 }
 
-//-----------------------------------------------------------------------------
+//------------------------------------------------------------------------------
 
 static int MyWriter(const uint8_t* data, size_t data_size,
                     const WebPPicture* const pic) {
@@ -533,7 +615,8 @@ static int DumpPicture(const WebPPicture* const picture, const char* PGM_name) {
   const int uv_width = (picture->width + 1) / 2;
   const int uv_height = (picture->height + 1) / 2;
   const int stride = (picture->width + 1) & ~1;
-  const int height = picture->height + uv_height;
+  const int alpha_height = picture->a ? picture->height : 0;
+  const int height = picture->height + uv_height + alpha_height;
   FILE* const f = fopen(PGM_name, "wb");
   if (!f) return 0;
   fprintf(f, "P5\n%d %d\n255\n", stride, height);
@@ -548,11 +631,16 @@ static int DumpPicture(const WebPPicture* const picture, const char* PGM_name) {
     if (fwrite(picture->v + y * picture->uv_stride, uv_width, 1, f) != 1)
       return 0;
   }
+  for (y = 0; y < alpha_height; ++y) {
+    if (fwrite(picture->a + y * picture->a_stride, picture->width, 1, f) != 1)
+      return 0;
+    if (picture->width & 1) fputc(0, f);  // pad
+  }
   fclose(f);
   return 1;
 }
 
-//-----------------------------------------------------------------------------
+//------------------------------------------------------------------------------
 
 static void HelpShort(void) {
   printf("Usage:\n\n");
@@ -567,7 +655,7 @@ static void HelpLong(void) {
   printf(" cwebp [-preset <...>] [options] in_file [-o out_file]\n\n");
   printf("If input size (-s) for an image is not specified, "
          "it is assumed to be a PNG or JPEG file.\n");
-#ifdef _WIN32
+#ifdef HAVE_WINCODEC_H
   printf("Windows builds can take as input any of the files handled by WIC\n");
 #endif
   printf("options:\n");
@@ -581,6 +669,8 @@ static void HelpLong(void) {
   printf("\n");
   printf("  -m <int> ............... compression method (0=fast, 6=slowest)\n");
   printf("  -segments <int> ........ number of segments to use (1..4)\n");
+  printf("  -size <int> ............ Target size (in bytes)\n");
+  printf("  -psnr <float> .......... Target PSNR (in dB. typically: 42)\n");
   printf("\n");
   printf("  -s <int> <int> ......... Input size (width x height) for YUV\n");
   printf("  -sns <int> ............. Spatial Noise Shaping (0:off, 100:max)\n");
@@ -588,26 +678,60 @@ static void HelpLong(void) {
   printf("  -sharpness <int> ....... "
          "filter sharpness (0:most .. 7:least sharp)\n");
   printf("  -strong ................ use strong filter instead of simple.\n");
+  printf("  -partition_limit <int> . limit quality to fit the 512k limit on\n");
+  printf("                           "
+         "the first partition (0=no degradation ... 100=full)\n");
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+  printf("  -alpha_comp <int> ...... set the transparency-compression\n");
+  printf("  -noalpha ............... discard any transparency information.\n");
+#endif
   printf("  -pass <int> ............ analysis pass number (1..10)\n");
   printf("  -crop <x> <y> <w> <h> .. crop picture with the given rectangle\n");
+  printf("  -resize <w> <h> ........ resize picture (after any cropping)\n");
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+  printf("  -444 / -422 / -gray ..... Change colorspace\n");
+#endif
   printf("  -map <int> ............. print map of extra info.\n");
   printf("  -d <file.pgm> .......... dump the compressed output (PGM file).\n");
+
   printf("\n");
   printf("  -short ................. condense printed message\n");
   printf("  -quiet ................. don't print anything.\n");
   printf("  -version ............... print version number and exit.\n");
+#ifndef WEBP_DLL
+  printf("  -noasm ................. disable all assembly optimizations.\n");
+#endif
   printf("  -v ..................... verbose, e.g. print encoding/decoding "
          "times\n");
   printf("\n");
   printf("Experimental Options:\n");
-  printf("  -size <int> ............ Target size (in bytes)\n");
-  printf("  -psnr <float> .......... Target PSNR (in dB. typically: 42)\n");
   printf("  -af .................... auto-adjust filter strength.\n");
   printf("  -pre <int> ............. pre-processing filter\n");
   printf("\n");
 }
 
-//-----------------------------------------------------------------------------
+//------------------------------------------------------------------------------
+// Error messages
+
+static const char* const kErrorMessages[] = {
+  "OK",
+  "OUT_OF_MEMORY: Out of memory allocating objects",
+  "BITSTREAM_OUT_OF_MEMORY: Out of memory re-allocating byte buffer",
+  "NULL_PARAMETER: NULL parameter passed to function",
+  "INVALID_CONFIGURATION: configuration is invalid",
+  "BAD_DIMENSION: Bad picture dimension. Maximum width and height "
+  "allowed is 16383 pixels.",
+  "PARTITION0_OVERFLOW: Partition #0 is too big to fit 512k.\n"
+  "To reduce the size of this partition, try using less segments "
+  "with the -segments option, and eventually reduce the number of "
+  "header bits using -partition_limit. More details are available "
+  "in the manual (`man cwebp`)",
+  "PARTITION_OVERFLOW: Partition is too big to fit 16M",
+  "BAD_WRITE: Picture writer returned an I/O error"
+  "FILE_TOO_BIG: File would be too big to fit in 4G"
+};
+
+//------------------------------------------------------------------------------
 
 int main(int argc, const char *argv[]) {
   const char *in_file = NULL, *out_file = NULL, *dump_file = NULL;
@@ -615,12 +739,18 @@ int main(int argc, const char *argv[]) {
   int c;
   int short_output = 0;
   int quiet = 0;
+  int keep_alpha = 0;
   int crop = 0, crop_x = 0, crop_y = 0, crop_w = 0, crop_h = 0;
+  int resize_w = 0, resize_h = 0;
   WebPPicture picture;
   WebPConfig config;
   WebPAuxStats stats;
   Stopwatch stop_watch;
 
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+  keep_alpha = 1;
+#endif
+
   if (!WebPPictureInit(&picture) || !WebPConfigInit(&config)) {
     fprintf(stderr, "Error! Version mismatch!\n");
     goto Error;
@@ -651,18 +781,18 @@ int main(int argc, const char *argv[]) {
     } else if (!strcmp(argv[c], "-m") && c < argc - 1) {
       config.method = strtol(argv[++c], NULL, 0);
     } else if (!strcmp(argv[c], "-q") && c < argc - 1) {
-      config.quality = strtod(argv[++c], NULL);
+      config.quality = (float)strtod(argv[++c], NULL);
     } else if (!strcmp(argv[c], "-size") && c < argc - 1) {
       config.target_size = strtol(argv[++c], NULL, 0);
     } else if (!strcmp(argv[c], "-psnr") && c < argc - 1) {
-      config.target_PSNR = strtod(argv[++c], NULL);
+      config.target_PSNR = (float)strtod(argv[++c], NULL);
     } else if (!strcmp(argv[c], "-sns") && c < argc - 1) {
       config.sns_strength = strtol(argv[++c], NULL, 0);
     } else if (!strcmp(argv[c], "-f") && c < argc - 1) {
       config.filter_strength = strtol(argv[++c], NULL, 0);
     } else if (!strcmp(argv[c], "-af")) {
       config.autofilter = 1;
-    } else if (!strcmp(argv[c], "-strong") && c < argc - 1) {
+    } else if (!strcmp(argv[c], "-strong")) {
       config.filter_type = 1;
     } else if (!strcmp(argv[c], "-sharpness") && c < argc - 1) {
       config.filter_sharpness = strtol(argv[++c], NULL, 0);
@@ -672,14 +802,37 @@ int main(int argc, const char *argv[]) {
       config.preprocessing = strtol(argv[++c], NULL, 0);
     } else if (!strcmp(argv[c], "-segments") && c < argc - 1) {
       config.segments = strtol(argv[++c], NULL, 0);
+    } else if (!strcmp(argv[c], "-partition_limit") && c < argc - 1) {
+      config.partition_limit = strtol(argv[++c], NULL, 0);
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+    } else if (!strcmp(argv[c], "-alpha_comp") && c < argc - 1) {
+      config.alpha_compression = strtol(argv[++c], NULL, 0);
+    } else if (!strcmp(argv[c], "-noalpha")) {
+      keep_alpha = 0;
+#endif
     } else if (!strcmp(argv[c], "-map") && c < argc - 1) {
       picture.extra_info_type = strtol(argv[++c], NULL, 0);
+#ifdef WEBP_EXPERIMENTAL_FEATURES
+    } else if (!strcmp(argv[c], "-444")) {
+      picture.colorspace = WEBP_YUV444;
+    } else if (!strcmp(argv[c], "-422")) {
+      picture.colorspace = WEBP_YUV422;
+    } else if (!strcmp(argv[c], "-gray")) {
+      picture.colorspace = WEBP_YUV400;
+#endif
     } else if (!strcmp(argv[c], "-crop") && c < argc - 4) {
       crop = 1;
       crop_x = strtol(argv[++c], NULL, 0);
       crop_y = strtol(argv[++c], NULL, 0);
       crop_w = strtol(argv[++c], NULL, 0);
       crop_h = strtol(argv[++c], NULL, 0);
+    } else if (!strcmp(argv[c], "-resize") && c < argc - 2) {
+      resize_w = strtol(argv[++c], NULL, 0);
+      resize_h = strtol(argv[++c], NULL, 0);
+#ifndef WEBP_DLL
+    } else if (!strcmp(argv[c], "-noasm")) {
+      VP8GetCPUInfo = NULL;
+#endif
     } else if (!strcmp(argv[c], "-version")) {
       const int version = WebPGetEncoderVersion();
       printf("%d.%d.%d\n",
@@ -727,9 +880,10 @@ int main(int argc, const char *argv[]) {
   }
 
   // Read the input
-  if (verbose)
+  if (verbose) {
     StopwatchReadAndReset(&stop_watch);
-  if (!ReadPicture(in_file, &picture)) {
+  }
+  if (!ReadPicture(in_file, &picture, keep_alpha)) {
     fprintf(stderr, "Error! Cannot read input picture\n");
     goto Error;
   }
@@ -761,15 +915,26 @@ int main(int argc, const char *argv[]) {
   picture.stats = &stats;
 
   // Compress
-  if (verbose)
+  if (verbose) {
     StopwatchReadAndReset(&stop_watch);
+  }
   if (crop != 0 && !WebPPictureCrop(&picture, crop_x, crop_y, crop_w, crop_h)) {
     fprintf(stderr, "Error! Cannot crop picture\n");
     goto Error;
   }
-  if (picture.extra_info_type > 0) AllocExtraInfo(&picture);
+  if ((resize_w | resize_h) > 0) {
+    if (!WebPPictureRescale(&picture, resize_w, resize_h)) {
+      fprintf(stderr, "Error! Cannot resize picture\n");
+      goto Error;
+    }
+  }
+  if (picture.extra_info_type > 0) {
+    AllocExtraInfo(&picture);
+  }
   if (!WebPEncode(&config, &picture)) {
     fprintf(stderr, "Error! Cannot encode picture as WebP\n");
+    fprintf(stderr, "Error code: %d (%s)\n",
+            picture.error_code, kErrorMessages[picture.error_code]);
     goto Error;
   }
   if (verbose) {
@@ -778,8 +943,12 @@ int main(int argc, const char *argv[]) {
   }
 
   // Write info
-  if (dump_file) DumpPicture(&picture, dump_file);
-  if (!quiet) PrintExtraInfo(&picture, short_output);
+  if (dump_file) {
+    DumpPicture(&picture, dump_file);
+  }
+  if (!quiet) {
+    PrintExtraInfo(&picture, short_output);
+  }
 
  Error:
   free(picture.extra_info);
@@ -791,4 +960,4 @@ int main(int argc, const char *argv[]) {
   return 0;
 }
 
-//-----------------------------------------------------------------------------
+//------------------------------------------------------------------------------
